#include "srcguru.h"

#include "hash.h"

#include "symbols.h"

#include "eval.h"

/***************** Item in the expression *********************************/

typedef struct itm {
    char* item_text;
    ev* item_value;
    char item_op;
    struct itm* next;
} item;

/***************** Get next token from expression text ********************/

char* get_exp_token(char* data, char* out)
{
    int paren = 0;
    char* b = out;
    while (*data) {
        if (paren == 0) {
            if (*data == '+') break;
            if (*data == '-') break;
            if (*data == '*') break;
            if (*data == '/') break;
        }
        if (*data == '(') paren++;
        if (*data == ')') paren--;
        if (paren >= 0)
            *out++ = *data++;
    }
    *out = 0;
    return data;
}

/***************** Convert expression to string ***************************/

void exp_to_str(ev* expr, char* out)
{
    char temp[20];
    label* l;

    if (expr == NULL) {
        strcpy(out, "(null)");
        return;
    }

    sprintf(out, "%d", expr->num);
    l = expr->first_label;
    while (l) {
        sprintf(temp, "%+d*", l->mult);
        strcat(out, temp);
        strcat(out, l->name);
        l = l->next_label;
    }
}    

/***************** Evaluate a single item *********************************/

ev* eval_item(char* item)
{
    ev* expr;
    int x;
    hash_item* y;
    label* l;

    if (*item == '(') {                                 /* Sub-expression */
        item[strlen(item) - 1] = 0;
        return eval_expression(item + 1);
    }

    expr = (ev *) malloc(sizeof(ev));
    expr->num = 0;
    expr->first_label = NULL;

    if (isdigit(*item)) {                               /* Decimal number */
        expr->num = atoi(item);;
        return expr;
    }
    if (*item == '$') {                                     /* Hex number */
        sscanf(item + 1, "%x", &expr->num);
        return expr;
    }

    y = get_symbol(item);                                       /* Symbol */
    if (y == NULL)                  
        return NULL;

    if (y->type == ST_EQUATE) {
        free(expr);
        return eval_expression(y->value);
    } else {
        l = (label *) malloc(sizeof(label));
        l->mult = 1;
        l->name = (char *) malloc(strlen(item) + 1);
        strcpy(l->name, item);
        l->next_label = NULL;
        expr->first_label = l;
        return expr;
    }

    return NULL;
}

/***************** Clean up an expression *********************************/

void clean_up(ev* expr)
{
    label** x;
    x = &expr->first_label;
    while (*x)
        if (((*x)->mult) == 0)
            *x = (*x)->next_label;
        else
            x = &(*x)->next_label;
}

/***************** Locate a label in an expression value ******************/

label* locate_label(char* name, ev* expr)
{
    label* x;
    x = expr->first_label;
    while (x) 
        if (eq(x->name, name))
            return x;
        else
            x = x->next_label;
    return NULL;
}

/***************** Add the next item to this item *************************/

void add_items(item* i1)
{
    item* i2;
    ev* expr;
    ev* expr2;
    label* x;
    label* y;

    i2 = i1->next;
    expr = i1->item_value;
    expr2 = i2-> item_value;
    expr->num += expr2->num;
    i1->next = i2->next;
    i1->item_op = i2->item_op;

    free(i2);

    x = expr2->first_label;
    while (x) {
        if (y = locate_label(x->name, expr))
            y->mult += x->mult;
        else {
            y = (label *) malloc(sizeof(label));
            y->mult = x->mult;
            y->name= (char *) malloc(strlen(x->name) + 1);
            strcpy(y->name, x->name);
            y->next_label = expr->first_label;
            expr->first_label = y;
        }
        x = x->next_label;
    }

    clean_up(expr);
}

/***************** Scale an expression by a constant **********************/

void scale_expr(ev* expr, int x)
{
    label* y;
    expr->num *= x;
    y = expr->first_label;
    while (y) {
        y->mult *= x;
        y = y->next_label;
    }
}

/***************** Subtract the next item from this item ******************/


void sub_items(item* i1)
{
    scale_expr(i1->next->item_value, -1);
    add_items(i1);
}

/***************** Overall expression evaluator ***************************/

ev* eval_expression(char* expr_txt)
{
    char it[MAX_LINE];
    item* x;
    item* items;
    item* last;

    remove_blanks(expr_txt);

    /* Build list of items in expression */

    items = (item *) malloc(sizeof(item));
    x = items;

    if (expr_txt[0] == '-') 
        strcpy(it, "0");
    else
        expr_txt = get_exp_token(expr_txt, it);
    x->item_text = (char *) malloc(strlen(it) + 1);
    x->item_op = 0;
    x->next = NULL;
    last = x;
    strcpy(x->item_text, it);

    while (*expr_txt) {
        if (*(expr_txt + 1) == 0)
            break;
        x->item_op = *expr_txt++;

        x->next = (item *) malloc(sizeof(item));
        x = x->next;
        last = x;
        x->item_op = 0;
        expr_txt = get_exp_token(expr_txt, it);
        x->item_text = (char *) malloc(strlen(it) + 1);
        x->next = NULL;
        strcpy(x->item_text, it);
    }

    /* Evaluate each item of the expression */

    x = items;
    while (x) {
        x->item_value = eval_item(x->item_text);
        if (x->item_value == NULL)
            return NULL;
        free(x->item_text);
        x = x->next;
    }

    /* Evaluate the expression */

    x = items;
    while (x->next) 
        if (x->item_op == '*') {
            if (classify_expression(x->item_value) == ET_CONST) {
                scale_expr(x->next->item_value, x->item_value->num);
                x->item_value = x->next->item_value;
                x->item_op = x->next->item_op;
                x->next = x->next->next;
            } else if (classify_expression(x->next->item_value) == ET_CONST) {
                scale_expr(x->item_value, x->next->item_value->num);
                x->item_op = x->next->item_op;
                x->next = x->next->next;
            } else
                return NULL;
        } else
            x = x->next;

    x = items;
    while (x->next) 
        if (x->item_op == '+')
            add_items(x);
        else if (x->item_op == '-')
            sub_items(x);
        else
            x = x->next;            

    exp_to_str(items->item_value, it);

    return items->item_value;
}

/***************** Classify expressions ***********************************/

int classify_expression(ev* expr)
{
    int labels = 0;
    label* x;

    if (expr == NULL)
        return ET_ERROR;
    x = expr->first_label;
    while (x) {
        labels += x->mult;
        x = x->next_label;
    }
    if (labels == 0)
        return ET_CONST;
    if (labels == 1)
        return ET_LABEL;
    return ET_OTHER;
}
