#include <stdio.h>

#include "srcguru.h"

#include "hash.h"

#include "symbols.h"

#include "macros.h"

#include "parser.h"

char line[MAX_LINE];
char stack[MAX_LINE];
char* data;
int lines = 0;

/***************** Character type testing **********************************/

int is_blank(char a)
{
    return ((a == ' ') || (a == 9) || (a == 0));
}

int is_directive(char a)
{
    return ((a == '#') || (a == '.') || (a == 0));
}

int is_not_token(char a)
{
    return (is_blank(a) || (a == ','));
}

/***************** Retrieve a single token *********************************/

char* get_token(char* in, char* out)
{
    int p = 0;
    while (1) {
        if (*in == '(')
            p++;
        if (*in == ')')
            p--;
        if ((p == 0) && (is_not_token(*in)))
            break;
        if (*in == 0)
            break;
        if (!(is_blank(*in)))
            *out++ = *in;
        in++;
    }
    *out = 0;
    return in;
}

/***************** Retrieve the next token *********************************/

char* get_next_token(char* in, char* out)
{
    while (is_not_token(*in)) {
        if (*in == 0) {
            *out = 0;
            return in;
        }
        in++;
    }
    return get_token(in, out);
}

/***************** Find next non-blank character ***************************/

char* next_data(char* in)
{
    while (is_blank(*in)) {
        if (*in == 0)
            return NULL;
        in++;
    }
    return in;
}

/***************** Get next thing in comma-delimited list ******************/

char* get_next_comma_del(char* in, char* out)
{
    in = next_data(in);
    if (in == NULL) {
        *out = 0;
        return in;
    }

    while (*in) {
        if (*in == ',') {
            in++;
            break;
        }
        if (!(is_blank(*in)))
            *out++ = *in;
        in++;
    }
    *out = 0;
    return in;
}

/***************** Get next quoted thing in comma-delimited list ***********/

char* get_next_comma_del_q(char* in, char* out)
{
    in = next_data(in);
    if (in == NULL) {
        *out = 0;
        return in;
    }

    while (*in) {
        if (*in == '"') {
            *out++ = *in++;
            while (*in) {
                *out++ = *in;
                if (*in++ == '"')
                    break;
            }
        }
        if (*in == ',') {
            in++;
            break;
        }
        if (!(is_blank(*in)))
            *out++ = *in;
        in++;
    }
    *out = 0;
    return in;
}

/***************** Load the next line **************************************/

int read_line(FILE* srcfile)
{
    char* temp;
    int q = 0;

    if (feof(srcfile))
        return 0;
    if (fgets(line, MAX_LINE, srcfile) == NULL)
        return 0;

    temp = line;
    do {
        if (*temp == 10)
            *temp = 0;
        if (*temp == 13)
            *temp = 0;
    } while (*temp++);

    if (line[strlen(line) - 1] == '\n')
        line[strlen(line) - 1] = 0;
    data = line;
    add_item(ITEM_LINE, line, empty, empty);

    temp = line;
    while (*temp) {
        if (*temp == '"')
            q = (q + 1) % 2;
        if ((*temp == ';') && (q == 0)) {
            add_item(ITEM_COMMENT, temp, empty, empty);
            *temp = 0;
            break;
        }
        temp++;
    }           

    lines++;
    return 1;
}

/***************** Parse a directive ***************************************/

void parse_directive(char* directive)
{
    char text[MAX_LINE];
    char data[MAX_LINE];
    FILE* nextfile;

    directive = get_token(directive, text);
    if (eq(text, "include")) {
        directive = get_next_token(directive, data);
        if (data != NULL) {
            strcpy(text, data);
            if (*data == '"') {
                data[strlen(data) - 1] = 0;
                strcpy(text, data + 1);
            }
            if (nextfile = fopen(text, "r")) {
                do_parse_file(nextfile);
                fclose(nextfile);
            } else {
                fprintf(outfile, "******** WARNING: Unable to open include file %s\n", text);
                printf("******** WARNING: Unable to open include file %s\n", text);
            }
        }
    } else if (eq(text, "dw")) {
        while (1) {
            directive = get_next_token(directive, data);
            if (*data == 0)
                break;
            add_item(ITEM_CONSTWORD, data, empty, empty);
        }
    } else if (eq(text, "db")) {
        while (1) {
            directive = get_next_comma_del_q(directive, data);
            if (*data == 0)
                break;
            if (*data == '"') {
                data[strlen(data) - 1] = 0;
                add_item(ITEM_CONSTSTR, data + 1, empty, empty);
            } else
                add_item(ITEM_CONSTNUM, data, empty, empty);
        }
    } else if (eq(text, "org")) {
    } else if (eq(text, "define")) {
        directive = next_data(directive);
        directive = get_token(directive, text);
        directive = next_data(directive);
        new_macro(text, directive);
    } else if (eq(text, "end")) {
    } else {
        fprintf(outfile, "******** WARNING: Unknown directive %s\n", text);
        printf("******** WARNING: Unknown directive %s\n", text);
    }
}

/***************** Parse an instruction ************************************

 This function parses an individual instruction.  It first test whether it
 is a directive in which case parse_directive() is called.  If it is a
 macro, the macro is decoded and parse_instructions() is then called on the
 resulting list.  Otherwise, it was a simple instruction which is simply
 added to the list. */

void parse_instruction(char* inst)
{
    char body[MAX_LINE] = "";
    char arg1[MAX_LINE] = "";
    char arg2[MAX_LINE] = "";

    inst = next_data(inst);

    if (inst == NULL)
        return;

    if (is_directive(*inst)) {
        parse_directive(inst + 1);
        return;
    }

    inst = get_token(inst, body);

    if (is_macro(body)) {
        decode_macro(body, arg1);
        parse_instructions(arg1);
        return;
    }

    if (*inst) {
        inst = get_next_comma_del(inst, arg1);
        get_next_comma_del(inst, arg2);
    }
    if (*body)
        add_item(ITEM_INSTRUCTION, body, arg1, arg2);
}

/***************** Parse a list of instructions ****************************

 This parses an instruction or list of instruction separated with '/'.  It
 will call parse_instruction() on each single instruction in the list.  It
 will be called recursively if an instruction is a macro. */

void parse_instructions(char* insts)
{
    char* tmp;
    tmp = insts;
    while (1) {
        if (*tmp == 0) {
            parse_instruction(insts);
            break;
        }
        if (*tmp == '\\') {
            *tmp = 0;
            parse_instruction(insts);
            parse_instructions(tmp + 1);
            break;
        }
        tmp++;
    }   
}

/***************** Overall parsing routine *********************************

 This is the routine which actually parses a file.  It reads one line at a
 time.  First, it determines whether the line begins as a label, equate,
 or directive; if so, it processes that accordingly.  Then, whatever reamins
 after that (or the whole line if the line beginning is blank) is considered
 an instruction list, to be parsed by parse_instructions().  This function
 will be called recursively if there are include files. */

int do_parse_file(FILE* srcfile)
{
    char body[MAX_LINE + 1];
    char arg1[MAX_LINE + 1];
    char arg2[MAX_LINE + 3];
    char* scan;
    char* scan2;
    int equate;
    FILE* nextfile;

    while (read_line(srcfile)) {
        if (!(is_blank(line[0]))) {
            if (is_directive(line[0])) {
                parse_directive(line + 1);
            } else {
                scan = get_token(line, body);
                scan2 = line;
                while (*++scan2)
                    if (*scan2 == ':')
                        break;
                if ((*scan2 == ':') && (!(is_blank(*(scan2 + 1))))) {
                    *scan2 = 0;
                    add_label(line);
                    parse_instructions(scan2 + 1);
                } else {
                    scan = next_data(scan);
                    if (scan == NULL) {
                        add_label(body);
                    } else {
                        if (*scan == '=')
                            add_equate(body, scan);
                        else {
                            add_label(body);
                            parse_instructions(scan);
                        }
                    }
                }
            }
        } else
            parse_instructions(line);
    }
    return lines;
}

/******************* Main parsing routine **********************************

 This routine is called once per program execution.  It initializes the
 symbol and hash tables, then calls the actual parsing routine. */

int parse_file()
{
    init_symbols();
    init_macros();
    return do_parse_file(infile);
}
